﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using wscf.lightblue;
using wscf.lightblue.Build;
using System.IO;

namespace cfbuild
{
	class Program
	{
		static int exitCode = 0;

		static void Main(string[] args)
		{
			Console.Out.WriteLine("Contract First Build.");
			WriteToFile(DateTime.Now.ToString(), false);

			Activity.Log += new LogHandler(OnLog);
			Activity.Error += new ErrorHandler(OnError);
			try
			{
				_cmdLine = new Arguments(args);

				bool showUsage = false;
				bool.TryParse(_cmdLine["?"], out showUsage);
				if (showUsage)
				{
					ShowUsage();
					return;
				}

				string rootDirectory = GetFullPathFromCmdArgument(new[] { "rootDirectory", "r" }, ".");
                string contractsDirectory = GetFullPathFromCmdArgument(new[] { "contractsDirectory", "c" }, "Contracts");
                string assembliesFolder = GetFullPathFromCmdArgument(new[] { "assembliesFolder", "a" }, "Assemblies");

				string searchPatterns = _cmdLine[new string[] { "searchPatterns", "s" }];
				if (searchPatterns == null)
				{
					searchPatterns = "*.wsdl;*.xsd";
				}

				string excludeFolders = _cmdLine[new string[] { "excludeFolders", "e" }];
				if (excludeFolders == null)
				{
					excludeFolders = "Web References;Service References";
				}

				string projectsToBuild = _cmdLine[new string[] { "projectsToBuild", "p" }];

				string referencedAssemblies = _cmdLine[new string[] { "referencedAssemblies", "f" }];

			    string assemblyPrefix = _cmdLine[new string[] { "assemblyPrefix", "x" }];

				bool continueOnError = false;
				bool.TryParse(_cmdLine["continueOnError"], out continueOnError);

				bool rebuild = false;
				bool.TryParse(_cmdLine["rebuild"], out rebuild);

				bool debug = false;
				bool.TryParse(_cmdLine["debug"], out debug);

				Builder b = new Builder
				{
					ExcludeFolders = excludeFolders.Replace(';', ','),
					Name = "Contract First Build",
					ProjectsToBuild = projectsToBuild != null ? new List<string>(projectsToBuild.Split(';')) : null,
					RootDirectory = rootDirectory,
					SearchPatterns = searchPatterns.Replace(';', ','),
					ServiceAssembliesFolder = assembliesFolder,
					ServiceAssemblyPrefix = assemblyPrefix,
					ServiceContractsFolder = contractsDirectory,
					ContinueOnError = continueOnError,
					Rebuild = rebuild,
					Debug = debug,
					ReferencedAssemblies = referencedAssemblies != null ? new List<string>(referencedAssemblies.Split(';')) : null,
				};

				b.Execute();
				if (b.IsError)
				{
					OnError(null, null, null);
				}
			}
			catch (Exception exc)
			{
				OnError(null, exc.ToString(), exc);
			}
			finally
			{
				Activity.Log -= new LogHandler(OnLog);
				Activity.Error -= new ErrorHandler(OnError);
			}

			Console.Out.WriteLine();
			if (exitCode == 0)
			{
				Console.Out.WriteLine("Build completed successfully!");
			}
			else
			{
				Console.Out.WriteLine("Build failed!");
			}
			Console.Out.WriteLine("Check build log for details:\n" + GetBuildLogFile());

			Environment.Exit(exitCode);
		}

     private static string GetFullPathFromCmdArgument(string[] argumentNames, string defaultPath)
	    {
	        string argumentValue = _cmdLine[argumentNames];
	        string path = string.IsNullOrWhiteSpace(argumentValue) ? defaultPath : argumentValue;
            return Path.GetFullPath(path);
	    }

	    private static string GetBuildLogFile()
		{
			return Path.Combine(
				Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData),
				"cfbuild.txt");
		}

		static private void WriteToFile(string s, bool append)
		{
			using (TextWriter tw = new StreamWriter(GetBuildLogFile(), append))
			{
				tw.WriteLine(s);
			}
		}

		static private int dots = 0;
		static private char[] dotsigns = new char[] { '.', '|' };
		static private int sign = 0;
	    private static Arguments _cmdLine;

	    static private void OnLog(object sender, string message)
		{
			WriteToFile(message, true);
			Console.Out.Write(dotsigns[sign]);
			if ((++dots % 20) == 0)
			{
				Console.Out.Write("\r");
				sign = sign == 0 ? 1 : 0;
			}
		}

		static private void OnError(object sender, string message, Exception exc)
		{
			if (!string.IsNullOrEmpty(message))
			{
				WriteToFile(message, true);
				Console.Out.WriteLine();
				Console.Out.WriteLine(message);
				Console.Out.WriteLine();
			}
			exitCode = 1;
		}

		static private void ShowUsage()
		{
			Console.Out.WriteLine(@"

Usage: cfbuild [options]

Options:

/rootDirectory (/r): The path to the root folder of the projects structure that contains service contracts. 
Contracts will be copied to contractsDirectory and service assemblies will be generated for each WSDL file.

/contractsDirectory (/c): Drop folder for service contracts. Deafult: "".\Contracts""

/searchPatterns (/s): List of search patterns separated by semicolon. Default: ""*.wsdl;*.xsd""

/excludeFolders (/e): List of folders to exclude in search for contracts separated by semicolon. Default: ""Web References;Service References""

/projectsToBuild (/p): List of projects or Visual Studio solutions that must be built in addtion to service assemblies. Separated by semicolon.

/referencedAssemblies (/f): List of assemblies to refence when generating service code and building service assemblies. Separated by semicolon.

/assembliesFolder (/a): Service assemblies drop folder

/assemblyPrefix (/x): Service assembly prefix, i.e. myorg.soa.

/continueOnError: Continue on error. Default: false

/rebuild: Clean contract and assembly folders and rebuild all. Default: false

/configuration (/g): Build configuration Release/Debug. Default: Release

Example: 

cfbuild /r:..\.. /c:..\Contracts /s:*.wsdl;*xsd /p:..\Common\Common.csproj;..\Data\Data.csproj;..\Parameter\Parameter.csproj /a:..\Assemblies /f:Common.dll;Data.dll;Parameter.dll");
		}
	}
}
