﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.ComponentModel;

namespace wscf.lightblue.Build
{
	public class Builder : Sequence
	{
		[DefaultValue("Assemblies")]
		public string ServiceAssembliesFolder { get; set; }

		public string ServiceAssemblyPrefix { get; set; }
		
		public List<string> ProjectsToBuild { get; set; }

		public List<string> ReferencedAssemblies { get; set; }

		[DefaultValue(".")]
		public string RootDirectory { get; set; }

		[DefaultValue("Contracts")]
		public string ServiceContractsFolder { get; set; }

		[DefaultValue(false)]
		public bool Rebuild { get; set; }

		[DefaultValue("false")]
		public bool Debug { get; set; }

		[DefaultValue("*.wsdl,*.xsd")]
		public string SearchPatterns { get; set; }

		[DefaultValue("Web References,Service References")]
		public string ExcludeFolders { get; set; }

		public override void Execute()
		{
			List<string> contracts = new List<string>();

			Activity cleanTempFolder = new CleanFolder
			{
				Name = "Clean Build Temporary Folder",
				Folder = GetBuildTempFolder()
			};

			Activities = new List<Activity>();

			if (Rebuild)
			{
				Activities.Add(new CleanFolder
				{
					Name = "Clean Service Contracts Folder",
					Folder = ServiceContractsFolder
				});

				Activities.Add(new CleanFolder
				{
					Name = "Clean Service Assemblies Folder",
					Folder = ServiceAssembliesFolder
				});
			}

			Activities.AddRange(new List<Activity>
			{
				cleanTempFolder,
				new CopyServiceContracts
				{
					Name = "Copy Service Contracts",
					ExcludeFolders = this.ExcludeFolders,
					RootDirectory = this.RootDirectory,
					Files = contracts,
					SearchPatterns = this.SearchPatterns,
					ServiceAssembliesFolder = this.ServiceAssembliesFolder,
					ServiceAssemblyPrefix = this.ServiceAssemblyPrefix,
					ServiceContractsDirectory = this.ServiceContractsFolder
				},
				GetBuildCommonAssembliesActivity(),
				cleanTempFolder,
				new GenerateServiceContractCode
				{
					Name = "Generate Service Contract Code",
					ReferencedAssemblies = ReferencedAssemblies,
					ServiceAssembliesFolder = this.ServiceAssembliesFolder,
					ServiceContracts = contracts,
					ContinueOnError = this.ContinueOnError
				},
				new BuildServiceAssemblies
				{
					Name = "Build Service Contract Assemblies",
					Folder = GetBuildTempFolder(),
					ReferencedAssemblies = ReferencedAssemblies,
					SearchPattern = "*.cs",
					ServiceAssembliesFolder = this.ServiceAssembliesFolder,
					ServiceAssemblyPrefix = this.ServiceAssemblyPrefix,
					Debug = this.Debug
				},
				new XCopy
				{
					CopyOnlyIfSourceIsNewer = true,
					DestinationFolder = ServiceAssembliesFolder,
					FilesPattern = "*.dll",
					OverwriteExistingFiles = true,
					SourceFolder = GetBuildTempFolder()
				},
				new XCopy
				{
					CopyOnlyIfSourceIsNewer = true,
					DestinationFolder = ServiceAssembliesFolder,
					FilesPattern = "*.pdb",
					OverwriteExistingFiles = true,
					SourceFolder = GetBuildTempFolder()
				}
			});

			base.Execute();

			if (!IsError)
			{
				OnLog("Build completed successfully!");
			}
		}

		private Sequence GetBuildCommonAssembliesActivity()
		{
			Sequence s = new Sequence
			{
				Name = "Build Common Assemblies"
			};
			foreach (string project in ProjectsToBuild)
			{
				s.Activities.Add(new InvokeVisualStudioTool
				{
					Name = "Build Common Project",
					FileName = "msbuild",
					Arguments = string.Format("\"{0}\" /p:Configuration={1};OutDir={2}\\",
						project,
						Debug ? "Debug" : "Release",
						GetBuildTempFolder())
				});
				s.Activities.Add(new XCopy
				{
					CopyOnlyIfSourceIsNewer = true,
					DestinationFolder = ServiceAssembliesFolder,
					FilesPattern = "*.dll",
					OverwriteExistingFiles = true,
					SourceFolder = GetBuildTempFolder()
				});
				s.Activities.Add(new XCopy
				{
					CopyOnlyIfSourceIsNewer = true,
					DestinationFolder = ServiceAssembliesFolder,
					FilesPattern = "*.pdb",
					OverwriteExistingFiles = true,
					SourceFolder = GetBuildTempFolder()
				});
			}
			return s;
		}

		public override string ToString()
		{
			return string.Format(@"{0}
	ContinueOnError = {1}
	Debug = {2}
	ExcludeFolders = {3}
	ProjectsToBuild = {4}
	Rebuild = {5}
	ReferencedAssemblies = {6}
	RootDirectory = {7}
	SearchPatterns = {8}
	ServiceAssembliesFolder = {9}
	ServiceContractsFolder = {10}",
				base.ToString(),
				this.ContinueOnError,
				this.Debug,
				this.ExcludeFolders,
				this.ProjectsToBuild != null ? string.Join("\n\t\t", this.ProjectsToBuild) : null,
				this.Rebuild,
				this.ReferencedAssemblies != null ? string.Join("\n\t\t", this.ReferencedAssemblies) : null,
				this.RootDirectory,
				this.SearchPatterns,
				this.ServiceAssembliesFolder,
				this.ServiceContractsFolder
			);
		}

	}
}
