﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Diagnostics;
using System.Text.RegularExpressions;
using System.Xml;
using System.ComponentModel;

namespace wscf.lightblue.Build
{

	public class CopyServiceContracts : Activity
	{
		/// <summary> 
		/// Root Directory 
		/// </summary> 
		public string RootDirectory { get; set; }

		/// <summary> 
		/// Service Contract Directory 
		/// </summary> 
		public string ServiceContractsDirectory { get; set; }

		public string ServiceAssembliesFolder { get; set; }

		public string ServiceAssemblyPrefix { get; set; }

		/// <summary> 
		/// Contract Search Patterns, i.e. "*.wsdl,*.xsd"
		/// </summary> 
		public string SearchPatterns { get; set; }

		/// <summary> 
		/// Exclude Folders, i.e. "Service References,Web references,Test Results"
		/// </summary> 
		public string ExcludeFolders { get; set; }

		public List<string> Files { get; set; }

		public override void Execute()
		{
			base.Execute();
			if (Files == null)
			{
				Files = new List<string>();
			}
			OnLog(ToString());
			try
			{
				LogIndent();
				RecurseFolders(RootDirectory);
			}
			finally
			{
				LogUnIndent();
			}
		}

		private void RecurseFolders(string folder)
		{
			if (!ExcludeFolders.Split(',').Contains<string>(Path.GetFileName(folder))
				&& Path.GetFullPath(folder).ToLower() != Path.GetFullPath(ServiceContractsDirectory).ToLower())
			{
				OnFolder(folder);
				foreach (string subfolder in Directory.EnumerateDirectories(folder))
				{
					RecurseFolders(subfolder);
				}
			}
		}

		private void OnFolder(string folder)
		{
			foreach (string searchPattern in SearchPatterns.Split(','))
			{
				foreach (string file in Directory.EnumerateFiles(folder, searchPattern))
				{
					string sourceFile = file;
					string targetFile = Path.Combine(ServiceContractsDirectory, Path.GetFileName(file));
					sourceFile = Path.Combine(GetBuildTempFolder(), Path.GetFileName(file));
					File.Copy(file, sourceFile, true);
					FileInfo fi = new FileInfo(sourceFile);
					fi.IsReadOnly = false;
					RemoveRelativePathsFromImports(sourceFile);
					bool isWsdl = Path.GetExtension(sourceFile).ToLower() == ".wsdl";
					bool assemblyExist = File.Exists(string.Format(@"{0}\{1}{2}", 
						ServiceAssembliesFolder, 
						ServiceAssemblyPrefix,
						Path.ChangeExtension(Path.GetFileName(sourceFile), 
						".dll")));
					if (!File.Exists(targetFile) 
						|| !FilesAreEqual(sourceFile, targetFile)
						|| (isWsdl && assemblyExist == false))
					{
						OnLog(string.Format("{0} ({1})", Path.GetFileName(targetFile), Path.GetDirectoryName(sourceFile)));
						if (isWsdl && !Files.Contains(targetFile))
						{
							Files.Add(targetFile);
						}
						if (File.Exists(targetFile))
						{
							fi = new FileInfo(targetFile);
							fi.IsReadOnly = false;
						}
						File.Copy(sourceFile, targetFile, true);
					}
				}
			}
		}

		const int BYTES_TO_READ = sizeof(Int64);

		static bool FilesAreEqual(string file1, string file2)
		{
			FileInfo first = new FileInfo(file1), second = new FileInfo(file2);
			if (first.Length != second.Length)
				return false;

			int iterations = (int)Math.Ceiling((double)first.Length / BYTES_TO_READ);

			using (FileStream fs1 = first.OpenRead())
			using (FileStream fs2 = second.OpenRead())
			{
				byte[] one = new byte[BYTES_TO_READ];
				byte[] two = new byte[BYTES_TO_READ];

				for (int i = 0; i < iterations; i++)
				{
					fs1.Read(one, 0, BYTES_TO_READ);
					fs2.Read(two, 0, BYTES_TO_READ);

					if (BitConverter.ToInt64(one, 0) != BitConverter.ToInt64(two, 0))
						return false;
				}
			}

			return true;
		}
		
		private void RemoveRelativePathsFromImports(string file)
		{
			FileInfo fi = new FileInfo(file);
			fi.IsReadOnly = false;
			XmlDocument doc = new XmlDocument();
			doc.Load(file);
			foreach (XmlNode n in doc.SelectNodes("//*[local-name()='import']"))
			{
				XmlAttribute attr = n.Attributes["schemaLocation"];
				if (attr != null)
				{
					string schemaLocation = attr.Value;
					attr.Value = Path.GetFileName(schemaLocation);
				}
			}
            foreach (XmlNode n in doc.SelectNodes("//*[local-name()='include']"))
            {
                XmlAttribute attr = n.Attributes["schemaLocation"];
                if (attr != null)
                {
                    string schemaLocation = attr.Value;
                    attr.Value = Path.GetFileName(schemaLocation);
                }
            }
			doc.Save(file);
		}

		public override string ToString()
		{
			return string.Format("{0} [RootDirectory: {1}, ServiceContractsDirectory: {2}, SearchPatterns: {3}, ExcludeFolders: {4}]",
				base.ToString(), this.RootDirectory, this.ServiceContractsDirectory, this.SearchPatterns, this.ExcludeFolders);
		}
	}
}
